////////////////////////////////////////////////////////
// PROJECT     : MONSOON
// TITLE       : SEQUENCER MPU ASSEMBLER 
// VERSION     : 005
// DATE        : 10/10/2006
// AUTHOR      : PETER MOORE
// HISTORY     : 04/10/2003  ORIGINAL VERSION - KAVI CHOPRA.
//             : 02/16/2004  Changed output format to suit IPX configuration.
//             :             Cleaned up the code, added dump option, etc.
//             : 02/16/2004  Added symbolic constants for pattern memory
//             : 11/20/2004  Added boolean psuedo functions to assembler
//             : 10/10/2006  Corrected line number accounting, fixed problems with LPB parsing
//             :             Corrected problem with labels without instructions. Added assembler
//             :             psuedo instruction for 'include' and 'literal' 
//
// DESCRIPTION : CONVERTS AN ASSEMBLER SOURCE CODE TO BINARY EXECUTABLE FOR THE
//             : MONSOON SEQUENCER FPGA MPU.
//             :
// RESOURCES   :
////////////////////////////////////////////////////////

#include "stdio.h"
#include "string.h"
#include "util.h"
#include "math.h"
#include "time.h"
#include "ctype.h"
//
// Opcodes for Sequencer MPU instruction set.
#define LPP 0x01				// load pattern pointer 
#define	IPP 0x02				// Increment pattern pointer
#define	DPP 0x03				// Decrement pattern pointer
#define LDA 0x04				// Load device address register
#define LMR 0x05				// Load mode register
#define LSR 0x06				// Load board select register 
#define CAL 0x07				// Call subroutine
#define RET 0x08				// Return from subroutine
#define JCB 0x09				// Conditional direct jump 
#define LRB 0x0A				// Loop on register begin
#define LPB 0x0B				// Loop on pattern data begin
#define LPE 0x0C				// Loop end 
#define DMS 0x0D				// Delay milliseconds
#define DSC 0x0E				// Delay sequencer clock cycles
#define DUS 0x0F				// Delay microseconds
#define NOP 0x00				// No Operation - delay one sequencer clock period
// Psuedo Opcodes for assembler use.
#define INC 0xF6                // Psuedo INCLUDE instruction to include another file into the assembly
#define NOT 0xF7				// Psuedo negate instruction for DEFine statements
#define AND 0xF8				// Psuedo AND instruction for DEFine statements
#define OR  0xF9				// Psuedo OR instruction for DEFine statements
#define XOR 0xFA				// Psuedo XOR instruction for DEFine statements
#define JMP 0xFB				// Psuedo JMP instruction which is expanded into a JCB
#define ORG 0xFC				// Psuedo ORG directive
#define PAT 0xFD				// Directive used to identify symbolic addresses of pattern data in source code.
#define DEF 0xFE				// Directive used to define symbolic constants in program source code.
#define BAD	0xFF				// Used to denote a bad opcode value in the findOpcode routine 

// Default output format set to PPX 
#define PPXFMT	true
#define DHEDBG 	false
#define PATSEEK 0				// Used to determine the data type that is required when looking for a label when calling Write_Operands()
#define DEFSEEK 1				// PATSEEK looks for an address of a label in the pattern memory (patData list), 
#define PGMSEEK 2				// DEFSEEK looks for a constant value in the defData lists.
								// PGMSEEK looks for a jump label in the Labels list.

#define SRCLINELENGTH 121		// Source line maximum length (exluding comments). 
#define SRCOPLENGTH 81			// Maximum Operand string length.
#define PATMEMSIZE 2048			// Version 0 => 4 firmware provides 2048 x 16 bit data to the PAN, 1024 x 32 bit to the DHE
#define PGMMEMSIZE 1024			// Version 0 => 4 firmware provides 1024 x 16 bit data to the PAN, 4096 x 4 bit to the DHE
#define DEFDATASIZE 4096		// Number of defines that can be assigned
#define DHEPATMEMSIZE PATMEMSIZE / 2  // DHE memory limit which is 32 bit memory
#define DHEPGMMEMSIZE PGMMEMSIZE * 4  // DHE memory limit which is 4 bit memory

#define PGMSEGNAME "PGM"		// Used with ORG directive to define program segment
#define PATSEGNAME "PAT"		// Used with ORG directive to define pattern segment

#define DHEPGMSEGADDR 0x4000	// Default program segment address in DHE
#define DHEPATSEGADDR 0x1000	// Default pattern segment address in DHE

#define PRODUCTNAME "NOAO MONSOON ASSEMBLER"
#define PRODUCTVERS "5.0"

#define MAX_OPERANDS 16
#define MAX_INCLUDES 16

struct Label
{
	int Address;
	char String[SRCOPLENGTH];
} Label[PGMMEMSIZE];

struct PatternData
{
	char String[SRCOPLENGTH];
	char SValue[SRCOPLENGTH];
	unsigned long Value;
} patData[DHEPATMEMSIZE];

struct DefinedData
{
	char String[SRCOPLENGTH];
	char SValue[SRCLINELENGTH];
	unsigned long Value;
} defData[DEFDATASIZE];

unsigned char Mem[DHEPGMMEMSIZE + 6];	// The +6 allows for any overflow to occur gracefully

char AppendFile[SRCOPLENGTH];  // File name for append file option

char SourceStack[MAX_INCLUDES][SRCOPLENGTH];	// Include source filename stack
int LineNumStack[MAX_INCLUDES];					// Include Line Number stack
unsigned StackCntr = 0;							// The stack pointer

unsigned pgmAddr = 0;
unsigned patAddr = 0;
unsigned defAddr = 0;
unsigned MaxPgmAddr = 0;
unsigned MaxPatAddr = 0;
unsigned TotalLabels =0;
unsigned LabelWidth = 10;				// Eases formating for dmp output, sets default label string size
unsigned WarnCount = 0;
unsigned ErrCount = 0;
int		LineNum = 0;
bool	OpOpt = PPXFMT;					// Switch to allow output format for dheHdwDbg .ucd file compatibility.
bool	OpDmp = false;					// Switch to allow line dump of parser to stdout.
bool    OpSlnt = false;					// Switch to allow silent running i.e. no output to stdout.

//
// Sub-routine definitions
//
void usage();
int RunPassOne(FILE *);
int RunPassTwo(FILE *);
int Read_Line(char *, FILE *);
int Write_Operands( unsigned, char *, int, int);
int parseOptions(char *);
unsigned char findOpcode(char *);
unsigned long parseValue(char *);
void printDheDbg(FILE *, char *);
void printPpx(FILE *, char *);
int PushSource(char *);
int PopSource(void);

//
// Begin asm5.cpp source
//
main ( long argc, char *argv[] )
{
	FILE  *fps = NULL;
	FILE  *fpt = NULL;
	FILE  *fpa = NULL;
	char  *Dest;
	char  c;
	unsigned i, p;
	*AppendFile = '\0';

	if (argc < 3) {
		usage();
		goto bad_exit;
	}
	else if (argc > 3) {
		p = 1;
		while (argv[p][0] == '-')		// Parse out all the options
			parseOptions(argv[p++]);

		if ((fps = fopen(argv[p],"r")) == NULL) {		// Open the source file
			if(!OpSlnt)
				fprintf(stdout, "\nERROR:- Cannot open source file %s",argv[p]);
			ErrCount++;
			goto bad_exit;
		}
		if ((fpt = fopen(argv[p+1],"w")) == NULL) {		// Open the destination file
			if(!OpSlnt)
				fprintf(stdout,"\nERROR:- Cannot open target file %s",argv[p+1]);
			ErrCount++;
			goto bad_exit;
		}
		strcpy(SourceStack[0], argv[p]);
		Dest = argv[p+1];
	}
	else {
		if ((fps = fopen(argv[1],"r")) == NULL) {
			if(!OpSlnt)
				fprintf(stdout,"\nERROR:- Cannot open source file %s",argv[1]);
			ErrCount++;
			goto bad_exit;
		}
		if ((fpt = fopen(argv[2],"w")) == NULL) {
			if(!OpSlnt)
				fprintf(stdout,"\nERROR:- Cannot open target file %s",argv[2]);
			ErrCount++;
			goto bad_exit;
		}
		strcpy(SourceStack[0], argv[1]);
		Dest = argv[2];
	}
	
	if(!OpSlnt)
		fprintf(stdout,"\nPass 1 of assembler");

// Run pass one of the assembler
    if (RunPassOne(fps) < 0)
		goto bad_exit;

	if(!OpSlnt)
		fprintf(stdout,"\n\nPass 2 of assembler");

// Output the definition and symbolic address tables
	if(OpDmp) {
		if(defAddr != 0) {
			fprintf(stdout,"\n\n Defined constants\n%-*s VALUE", LabelWidth, "LABEL");
			for(i = 0; i < DEFDATASIZE ; i++) {
				if(strlen(defData[i].String) != 0)
					fprintf(stdout,"\n%-*s  0x%x", LabelWidth, defData[i].String, defData[i].Value);
			}
		}
		if(patAddr != 0) {
			fprintf(stdout,"\n\n Pattern memory locations\n%-*s PANADDR   VALUE", LabelWidth, "LABEL");
			for(i = 0; i < DHEPATMEMSIZE ; i++) {
				if(strlen(patData[i].String) != 0)
					fprintf(stdout,"\n%-*s  0x%4.4x   0x%8.8x", LabelWidth, patData[i].String, i * 2, patData[i].Value);
			}
		}
		fprintf(stdout,"\n\n Program label listing\n%-*s DHEADDR   LINE #", LabelWidth, "LABEL");
	}

// Re-open the source file (it would have been closed by the first pass subroutine)
	if ((fps = fopen(SourceStack[0],"r")) == NULL) {
		if(!OpSlnt)
			fprintf(stdout,"\nERROR:- Cannot open source file %s on pass 2",SourceStack[0]);
		ErrCount++;
		goto bad_exit;
	}
	rewind(fps);

	pgmAddr = 0;
	LineNum = 0;

	
// Run pass two to insert the program code symbolic jump and call addresses
    if (RunPassTwo(fps) < 0)
		goto bad_exit;
	

	if(!OpSlnt)
		fprintf(stdout,"\n\nWriting data out\n");

	switch (OpOpt)
	{
		case DHEDBG:
			printDheDbg(fpt, SourceStack[0]);
			break;
		case PPXFMT:
			printPpx(fpt, SourceStack[0]);
			break;
	}


// Add the append file if the option is set
	if(strlen(AppendFile) > 0) {
		if ((fpa = fopen(AppendFile,"r")) != NULL) {
			if(!OpSlnt)
				fprintf(stdout,"\nAppending file %s to %s\n", AppendFile, Dest);
			while((c = getc(fpa)) != EOF)
				putc(c, fpt);

		}
		else {
			if(!OpSlnt)
				fprintf(stdout,"\nWARNING:- Could not open append file %s !",AppendFile);
			WarnCount++;
		}	
	}

	bad_exit:

// Clean up and release resources
	if(fpt != NULL)
		fclose(fpt);

	if(!OpSlnt) {
		fprintf(stdout,"\nUsage: %d (%d%%) locations used in Pattern segment. %d (%d%%) locations used in Program segment.", MaxPatAddr, (int)ceil(((double)MaxPatAddr) / ((double)DHEPATMEMSIZE) * 100.0), MaxPgmAddr, (int)ceil(((double)MaxPgmAddr) / ((double)DHEPGMMEMSIZE) * 100.0));  
		fprintf(stdout,"\nResources: %d locations avail in Pattern segment. %d locations avail in Program segment.", DHEPATMEMSIZE, DHEPGMMEMSIZE);  
		fprintf(stdout,"\n %d Warning   %d Errors -- Done !\n", WarnCount, ErrCount);
	}

// Just how good was the code ?
	if(ErrCount != 0 || WarnCount != 0)
		return(-1);

	return(0);
}


//
// Assembler First Pass loop
//
int RunPassOne(FILE *fps)
{

	FILE  *kps = NULL;
	
	char  *Line ;
	char  *Lbl;
	char  Opcode[SRCOPLENGTH];
	char  Operands[MAX_OPERANDS][SRCOPLENGTH];

	unsigned char g;
	unsigned q, defAccum;
	bool notPending, andPending, orPending, xorPending;
	unsigned i = 0, j = 0, orgPat, orgPgm;

	int errnum = 0;

	Line = (char *)malloc(SRCLINELENGTH);
	Lbl =  (char *)malloc(SRCOPLENGTH);	

	
	while((Read_Line(Line,fps) != EOF))
	{	

		if(OpDmp)
			fprintf(stdout,"\n%3.3d %s",LineNum, Line);

        // Clear the operands array
		for(q = 0; q < MAX_OPERANDS;q++)
			Operands[q][0]='\0';
		Opcode[0] = '\0';


		if (*Line == '#') {
			sscanf(Line,"%s %s %s %s %s",Lbl, Opcode, &Operands[0], &Operands[1], &Operands[2]); 
			strcpy(Label[TotalLabels].String,Lbl);
			LabelWidth = max(LabelWidth, strlen(Lbl));			// Measure the max extent of the label name
			Label[TotalLabels++].Address = pgmAddr;
		}
		else			
			sscanf(Line,"%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s",Opcode, 
			       &Operands[0], &Operands[1], &Operands[2], &Operands[3], 
			       &Operands[4], &Operands[5], &Operands[6], &Operands[7], 
			       &Operands[8], &Operands[9], &Operands[10], &Operands[11], 
			       &Operands[12], &Operands[13], &Operands[14], &Operands[15]); 
					
		if (strlen(Opcode) < 3)
			continue;                           // Escape for lines with just a symbolic address

		g = findOpcode(Opcode);
		
		switch (g) {
			case IPP:							// Increment pattern pointer
			case DPP:							// Decrement pattern pointer
			case RET:							// Return from subroutine
			case LPB:							// Loop on pattern memory data begin
			case LPE:							// Loop end 
			case NOP:							// No Operation
				Mem[pgmAddr] = g;
				pgmAddr = pgmAddr + 1;
				break;
			case LMR:							// Load mode register
			case LRB:							// Loop on register begin
				Mem[pgmAddr] = g;
				if(Write_Operands(1,Operands[0], LineNum, DEFSEEK) < 0) {
					errnum = -1;
					goto bad_exit_1;
				}
				pgmAddr = pgmAddr + 2;
				break;
			case LDA:							// Load device address register
			case LSR:							// Load board select register
			case DMS:							// Delay milliseconds
			case DUS:							// Delay microseconds
			case DSC:							// Delay sequencer clock
				Mem[pgmAddr] = g;
				if(Write_Operands(2,Operands[0], LineNum, DEFSEEK) < 0) {
					errnum = -1;
					goto bad_exit_1;
				}
				pgmAddr = pgmAddr + 3;
				break;
			case LPP:							// load pattern pointer
				Mem[pgmAddr] = g;
				if(Write_Operands(3,Operands[0], LineNum, PATSEEK) < 0) {
					errnum = -1;
					goto bad_exit_1;
				}
				pgmAddr = pgmAddr + 4;
				break;
			case CAL:							// Call subroutine 
				Mem[pgmAddr] = g;
				pgmAddr = pgmAddr + 4;		// Call address resolved in pass 2
				break;
			case JMP:							// Psuedo JMP instruction. Just stuffs a 0 in the operand1 value.
				strcpy(Operands[1],"0");
				g = JCB;                        // and fall through to the real instruction
			case JCB:							// Conditional direct jump 
				Mem[pgmAddr] = g;
				pgmAddr = pgmAddr + 3;		// Call address resolved in pass 2
				if(Write_Operands(1, Operands[1], LineNum, DEFSEEK) < 0) {
					errnum = -1;
					goto bad_exit_1;
				}
				pgmAddr = pgmAddr + 2;
				break;
			case PAT:							// Define pattern memory constant
				if(strlen(Operands[2]) != 0) {
					i = (unsigned) parseValue(Operands[2]) / 2;		// !!!! Needs some work here !!!!!
					if(strlen(patData[i].String) != 0) {
						if(!OpSlnt)
							fprintf(stdout,"\nWARNING:- Absolute Pattern memory value %s will overwrite currently assigned label %s at address ox%4.4x on line %3.3d of file %s",Operands[0], patData[i].String, i, LineNum, SourceStack[0]);
						WarnCount++;
					}
					strcpy(patData[i].String,Operands[0]);
					strcpy(patData[i].SValue, Operands[1]);
					if(*Operands[1] == '#') {
						for(j = 0; j < DEFDATASIZE; j++) {
							if(strcmp(Operands[1], defData[j].String) == 0) {
								patData[i].Value = defData[j].Value;
								break;
							}
						}
						if(j >= DEFDATASIZE) {
							if(!OpSlnt)
								fprintf(stdout,"\nERROR:- Definition for %s is not defined on line %3.3d in file %s", Operands[1], LineNum, SourceStack[0]);
							errnum = -1;
							ErrCount++;
							goto bad_exit_1;
						}
					}
					else
						patData[i].Value = parseValue(patData[i].SValue);
					LabelWidth = max(LabelWidth, strlen(Operands[0]));
				}
				else {
					while(strlen(patData[patAddr].String) != 0 && patAddr < DHEPATMEMSIZE)			// Skip over prior assignments made with absolute address
						patAddr++;
					if(patAddr >= DHEPATMEMSIZE) {
						if(!OpSlnt)
							fprintf(stdout,"\nERROR:- Available DHE Pattern memory segment space exhuasted !");
						ErrCount++;
						errnum = -1;
						goto bad_exit_1;
					}
					if (strlen(Operands[1]) > 0) {		// See if the pattern data has a leading symbol associated with it
						strcpy(patData[patAddr].String,Operands[0]);
						strcpy(patData[patAddr].SValue,Operands[1]);
					}
					else {
						*patData[patAddr].String = '\0';
						strcpy(patData[patAddr].SValue,Operands[0]);
					}
					if(*patData[patAddr].SValue == '#') {
						for(j = 0; j < DEFDATASIZE; j++) {
							if(strcmp(patData[patAddr].SValue, defData[j].String) == 0) {
								patData[patAddr].Value = defData[j].Value;
								break;
							}
						}
						if(j >= DEFDATASIZE) {
							if(!OpSlnt)
								fprintf(stdout,"\nWARNING:- Definition for %s is not defined on line %3.3d in file %s", patData[patAddr].SValue, LineNum, SourceStack[0]);
							WarnCount++;
						}
					}
					else
						patData[patAddr].Value = parseValue(patData[patAddr].SValue);
					LabelWidth = max(LabelWidth, strlen(Operands[0]));
					patAddr++;
				}
				break;
			case DEF:							// Define a constant
				if(defAddr >= DEFDATASIZE) {
					if(!OpSlnt)
						fprintf(stdout,"\nERROR:- Available Definition storage space exhuasted !");
					ErrCount++;
					errnum = -1;
					goto bad_exit_1;
				}
				strcpy(defData[defAddr].String,Operands[0]);
				defData[defAddr].Value = 0;
				notPending = andPending = orPending = xorPending = false;

				for(q=1;(*Operands[q] != 0) && (q < MAX_OPERANDS); q++) {
					strcat(defData[defAddr].SValue, Operands[q]);
					strcat(defData[defAddr].SValue, " ");
					g = findOpcode(Operands[q]);
					switch (g) {
						case NOT:
							notPending = !notPending;
							break;
						case AND:
							andPending = true;
							orPending = xorPending = false;
							break;
						case OR:
							orPending = true;
							andPending = xorPending = false;
							break;
						case XOR:
							xorPending = true;
							andPending = orPending = false;
							break;
						default:
							if(*Operands[q] == '#') {
								for(j = 0; j < DEFDATASIZE; j++) {
									if(strcmp(Operands[q], defData[j].String) == 0) {
										defAccum = defData[j].Value;
										break;
									}
								}
								if(j >= DEFDATASIZE) {
									if(!OpSlnt)
										fprintf(stdout,"\nWARNING:- Definition for %s is not defined on line %3.3din file %s", Operands[q], LineNum, SourceStack[0]);
									WarnCount++;
									break;
								}
							}
							else
								defAccum = parseValue(Operands[q]);

							if (notPending)
								defAccum = ~defAccum;
							if (andPending)
								defAccum = defData[defAddr].Value & defAccum;
							if (orPending)
								defAccum = defData[defAddr].Value | defAccum;
							if (xorPending)
								defAccum = defData[defAddr].Value ^ defAccum;
							defData[defAddr].Value = defAccum;
							andPending = orPending = xorPending = false;
							break;
					}
				}
				LabelWidth = max(LabelWidth, strlen(Operands[0]));
				defAddr++;
				break;
			case ORG:
				if(*Operands[1] == '#') {					// Extract the target value first
					for(j = 0; j < DEFDATASIZE; j++) {
						if(strcmp(Operands[1], defData[j].String) == 0) {
							orgPat = orgPgm = defData[j].Value;
							break;
						}
					}
					if(j >= DEFDATASIZE) {
						if(!OpSlnt)
							fprintf(stdout,"\nWARNING:- Definition for %s is not defined on line %3.3d in file %s", Operands[1], LineNum, SourceStack[0]);
						WarnCount++;
						orgPgm = pgmAddr;
						orgPat = patAddr * 2;
					}
				}
				else
					orgPat = orgPgm = parseValue(Operands[1]);

				if(strcmp(Operands[0], PGMSEGNAME) == 0) {	// Now identify for which segment
					if(orgPgm >= DHEPGMMEMSIZE) {
						if(!OpSlnt)
							fprintf(stdout,"\nWARNING:- ORG statement on line %3.3d in file %s will exceed allowed program memory space. Statement ignored.", LineNum, SourceStack[0]);
						WarnCount++;
						orgPgm = pgmAddr;
					}
					pgmAddr = orgPgm;					// Set the new origin
					if(*Line == '#') {
						TotalLabels--;
						Label[TotalLabels++].Address = pgmAddr;
					}
				}
				else
				if(strcmp(Operands[0], PATSEGNAME) == 0) {
					if(orgPat >= DHEPATMEMSIZE) {
						if(!OpSlnt)
							fprintf(stdout,"\nWARNING:- ORG statement on line %3.3d in file %s will exceed allowed pattern memory space. Statement ignored.", LineNum, SourceStack[0]);
						WarnCount++;
						orgPat = patAddr * 2;
					}
					patAddr = orgPat / 2;					// Set the new origin
				}
				break;
			case INC:
				if(strlen(Operands[0]) > 0) {
					if ((kps = fopen(Operands[0],"r")) != NULL) {						
						if(PushSource(Operands[0]) < 0) {
							if(!OpSlnt)
								fprintf(stdout,"\nERROR:- INCLUDE Level Overflow. Only 16 levels supported !");
							ErrCount++;
							errnum = -1;
							goto bad_exit_1;
						}
						if(OpDmp)
							fprintf(stdout,"\nIncluding File: %s",Operands[0]);
						if((errnum = RunPassOne(kps)) < 0)
							goto bad_exit_1;
					}
					else {
						if(!OpSlnt)
							fprintf(stdout,"\nERROR:- Could not open include file %s !",Operands[0]);
						ErrCount++;
						errnum = -1;
						goto bad_exit_1;
					}

				}
				else {
					if(!OpSlnt)
						fprintf(stdout,"\nWARNING:- INCLUDE File name missing in line %3.3d in file %s. Statement ignored.", LineNum, SourceStack[0]);
					WarnCount++;
				}
				break;
			default:
				if(!OpSlnt)
					fprintf(stdout,"\nWARNING:- Undefined SYNTAX ERROR: at LINE %3.3d in file %s",LineNum, SourceStack[0]);
				WarnCount++;
				break;

		}

		if(pgmAddr >= DHEPGMMEMSIZE) {				// Available program memory check.
			if(!OpSlnt)
				fprintf(stdout,"\nERROR:- Available DHE program segment space exhuasted !");
			ErrCount++;
			errnum = -1;
			goto bad_exit_1;
		}
		
		MaxPatAddr = max(MaxPatAddr, patAddr);

		*Line = '\0';
	}

	bad_exit_1:

// Clean up our local resources
	if(fps != NULL)
		fclose(fps);
	if(Line != NULL)
		free(Line);
	if(Lbl != NULL)
		free(Lbl);

// Try to recover the source file name
	PopSource();

	return(errnum);
}


//
// Assembler Pass Two Loop
//
int RunPassTwo (FILE *fps)
{

	FILE  *kps = NULL;

	char  *Line ;
	char  *Lbl;
	char  Opcode[SRCOPLENGTH];
	char  Operands[MAX_OPERANDS][SRCOPLENGTH];

	unsigned char g;
	unsigned q, i, j, orgPgm;

	int errnum = 0;

	Line = (char *)malloc(SRCLINELENGTH);
	Lbl =  (char *)malloc(SRCOPLENGTH);	

	
	while( (Read_Line(Line,fps) != EOF) )
	{

        // Clear the operands array
		for(q = 0; q < MAX_OPERANDS;q++)
			Operands[q][0]='\0';
		Opcode[0] = '\0';

		if (*Line == '#')
			sscanf(Line,"%s %s %s %s",Lbl, Opcode, Operands[0],Operands[1]) ;
		else			
			sscanf(Line,"%s %s %s",Opcode, Operands[0], Operands[1]);
		
		
		if (strlen(Opcode) < 3)
			continue;                           // Escape for lines with just a symbolic address

		g = findOpcode(Opcode);
		
		switch (g) {
			case IPP:
			case DPP:
			case RET:
			case LPB:
			case LPE:
			case NOP:
				pgmAddr = pgmAddr + 1;
				break;
			case LMR:
			case LRB:
				pgmAddr = pgmAddr + 2;
				break;
			case LDA:
			case LSR: 
			case DMS:
			case DUS:
			case DSC:
				pgmAddr = pgmAddr + 3;
				break;
			case LPP:
				pgmAddr = pgmAddr + 4;
				break;
			case CAL: 
				if((i = Write_Operands(3,Operands[0], LineNum, PGMSEEK)) < 0) {
					errnum = -1;
					goto bad_exit_2;
				}
				if(OpDmp)
					fprintf(stdout,"\n%-*s  0x%4.4x   %3.3d", LabelWidth, Label[i].String, Label[i].Address, LineNum);
				pgmAddr = pgmAddr + 4;
				break;
			case JMP:							// Psuedo JMP instruction. Just stuffs a 0 in the operand1 value.
			case JCB:
				if((i = Write_Operands(3,Operands[0], LineNum, PGMSEEK)) < 0) {
					errnum = -1;
					goto bad_exit_2;
				}
				if(OpDmp)
					fprintf(stdout,"\n%-*s  0x%4.4x   %3.3d", LabelWidth, Label[i].String, Label[i].Address, LineNum);
				pgmAddr = pgmAddr + 5;
				break;
			case ORG:
				break;
				if(strcmp(Operands[0], PGMSEGNAME) == 0) {		// identify for which segment, we're only interested in PGM
					if(*Operands[1] == '#') {					// Extract the target value
						for(j = 0; j < DEFDATASIZE; j++) {
							if(strcmp(Operands[1], defData[j].String) == 0) {
								orgPgm = defData[j].Value;
								break;
							}
						}
						if(j >= DEFDATASIZE)
							orgPgm = pgmAddr;
					}
					else
						orgPgm = parseValue(Operands[1]);

					if(orgPgm >= DHEPGMMEMSIZE)
						orgPgm = pgmAddr;
					pgmAddr = orgPgm;						// Set the new origin
				}
				break;
			case INC:
				if(strlen(Operands[0]) > 0) {
					if ((kps = fopen(Operands[0],"r")) != NULL) {
						PushSource(Operands[0]);
						if((errnum = RunPassTwo(kps)) < 0)
							goto bad_exit_2;
					}
					else {
						if(!OpSlnt)
							fprintf(stdout,"\nERROR:- Could not open include file %s !",Operands[0]);
						ErrCount++;
						errnum = -1;
						goto bad_exit_2;
					}
				}
				break;
			default:
				break;
		}
		
		MaxPgmAddr = max(MaxPgmAddr, pgmAddr);
	}

	bad_exit_2:

// Clean up our local resources
	if(fps != NULL)
		fclose(fps);
	if(Line != NULL)
		free(Line);
	if(Lbl != NULL)
		free(Lbl);

// Try to recover the source file name
	PopSource();

	return(errnum);
}


//
// Print out the assembler usage information
//
void usage()
{
	fprintf(stdout, "\n\nusage:  asm5 [-Option List] [Source file name] [Target file name]");
	fprintf(stdout, "\n        or");
	fprintf(stdout, "\n        asm5a [-Option] [-Option] [-Option] [Source file name] [Target file name]");
	fprintf(stdout, "\n\n        Available options are:-");
	fprintf(stdout, "\n        i Format output for dheDbgHdw configuration file.");
	fprintf(stdout, "\n        p Format output for PPX configuration tool. This is the default option");
	fprintf(stdout, "\n        d Dump line output while parsing, usefull for debugging assembler code");
	fprintf(stdout, "\n        s Silent assembly, no output to stdout. To use as a child process during waveform assy via make etc.");
	fprintf(stdout, "\n        a Append file to output file. Useful to concatinate micro sequencer code to the mcb mpu sequencer load file");
	fprintf(stdout, "\n          The filename must follow the a option without any space i.e. -aMyFileName.ucd");
	fprintf(stdout, "\n\n Sixteen native instructions + 1 psuedo instruction are supported." );
	fprintf(stdout, "\n Source Syntax :- [#LABEL] OPCODE [OPERAND1] [OPERAND2]");
	fprintf(stdout, "\n All labels must be prefixed by a '#'");
	fprintf(stdout, "\n The psuedo instruction 'JMP #LABEL' is available and is expanded to the native 'JCB #LABEL 0' form");
	fprintf(stdout, "\n\n Preprocessor directive 'PAT' (PATtern Image) is supported for pattern memory images");
	fprintf(stdout, "\n Syntax :- PAT [#LABEL] VALUE [ADDRESS]");
	fprintf(stdout, "\n Pattern memory values must be either a numeric operand or have been DEFined previously in the program code");
	fprintf(stdout, "\n The address field is optional and defines an absolute address from the PAN perspective. If the address field is");
	fprintf(stdout, "\n specified, the label field must also be present. Note; ORGing to an appropriate address is the preferred option");
	fprintf(stdout, "\n\n Preprocessor directive 'DEF' (DEFine constant) is supported for source code symbolic definitions");
	fprintf(stdout, "\n Syntax :- DEF #LABEL VALUE");
	fprintf(stdout, "\n Values must be DEFined before using them in the program code");
	fprintf(stdout, "\n\n The psuedo instructions 'AND','OR', 'NOT', and 'XOR' can be used up to 7 levels deep in DEF statements");
	fprintf(stdout, "\n Syntax :- DEF #LABEL VALUE OR VALUE AND VALUE AND NOT VALUE");
	fprintf(stdout, "\n Previously DEFined values can be used as the operands but grouping by parenthasis is not supported.");	
	fprintf(stdout, "\n\n Preprocessor directive 'ORG' (ORiGin memory pointer) is supported for pattern and program memory segments");
	fprintf(stdout, "\n Syntax :- [#LABEL] ORG SEGMENT ADDRESS - Segment definition must be either 'PGM' or 'PAT'");
	fprintf(stdout, "\n Note: The ORG PAT statement sets the pattern memory origin with respect");
	fprintf(stdout, "\n       to the PAN perspective but the ORG PGM sets the program origin with");
	fprintf(stdout, "\n       respect to the DHE perspective");
	fprintf(stdout, "\n\n Preprocessor directive 'INC' can be used to include another file within the assembly pass.");
	fprintf(stdout, "\n Syntax :- INC FILENAME - The file name can include a path. Important: The filename and path will");
	fprintf(stdout, "\n be interpreted using capital letters. This has implications for the LINUX user. Nested includes");
	fprintf(stdout, "\n are permitted. Maximum INClude depth is %d",MAX_INCLUDES - 1);
	fprintf(stdout, "\n\n A prefix can be used on numerical operands to specify the base in:-");
	fprintf(stdout, "\n    Decimal : (d)[Number]");
	fprintf(stdout, "\n    Hex     : (x/h)[Number]");
	fprintf(stdout, "\n    Binary  : (b)[Number]");
	fprintf(stdout, "\n\n Maximum source code line length is %d characters (excluding comments)", SRCLINELENGTH - 1);
	fprintf(stdout, "\n Maximum operand token length is %d characters.", SRCOPLENGTH - 1);
	fprintf(stdout, "\n");
	return;
}

//
// Parse the input options from command line
//
int parseOptions(char *c)
{
	int OpCnt = 0;

	if(*c++ == '-') {
		while(*c != '\0') {
			switch (*c) {
			case 'i':
				OpOpt = DHEDBG;
				OpCnt++;
				break;
			case 'p':
				OpOpt = PPXFMT;
				OpCnt++;
				break;
			case 'd':
				OpDmp = true;
				OpCnt++;
				break;
			case 's':
				OpDmp = false;
				OpSlnt = true;
				OpCnt++;
				break;
			case 'a':
				strcpy(AppendFile,++c);
				c = c + strlen(AppendFile) - 1;
				OpCnt++;
				break;
			default:
				if(!OpSlnt)
				WarnCount++;
				break;
			}
			c++;
		}
	}
	return(OpCnt);
}

//
// Read in a line from the input file and clean it up
//
int Read_Line(char *line, FILE * f)
{
    char c;
	bool commentFlag = false;
    bool startFlag = true;
	bool tabFlag = false;
	bool spaceFlag = false;

	while((c = getc(f)) != EOF)
	{

		if(c == '\r' || c == '\f')			// In case we're reading a UNIX file
			continue;

		if(c == '\n') {
			LineNum++;						// Account for the line whatever it is
			commentFlag = false;			// end of line terminates a comment
			if(startFlag == false)
				break;						// End of the line here
			continue;						// Or pass over null lines
		}
    
		if(c == '*') {						// Checks if line has a comment - skips the rest of the current line.
			commentFlag = true;
			continue;
		}

		if((c ==' ' || c == '\t') && startFlag == true)
			continue;						// Skips leading blanks and tabs

		if(commentFlag == true)
			continue;

		startFlag = false;

		if(c == '\t')
			c = ' ';						// replace the tabs with spaces

		if(c == ' ') {						// Strip out multiple (redundent) spaces
			if(spaceFlag == true)
				continue;
			spaceFlag = true;
		}
		else
			spaceFlag = false;

        *line++ = toupper(c);				// So we want this character - covert it to upper case
    }
	if(c == EOF)							// Catch the end of file
		return(c);
	
    line--; 
    while (*line == ' ' || *line == '\t')	// Rewinds to delete the terminating white spaces. 
        line--;
    line++;									// Inserts Null as a terminating character.
	*line = '\0';

    return(!EOF);
}


int Write_Operands(unsigned  OpSize, char * operand, int LineNum, int Seek)
{
	
	unsigned i,j;
	unsigned long m, l, value = 0;

	if(strlen(operand) == 0) {			// Just exit if there is nothing to work with
		if(!OpSlnt)
			fprintf(stdout,"\nERROR:- Operand missing from line %3.3d in file %s", LineNum, SourceStack[0]);
		ErrCount++;
		return(-1);
	}

	if(operand[0] == '#') {
		switch(Seek) {
			case PATSEEK:
				for(i = 0; i < DHEPATMEMSIZE; i++) {
					if(strcmp(operand, patData[i].String) == 0) {
						value = (unsigned long) i;
						break;
					}
				}
				if(i >= DHEPATMEMSIZE) {
					if(!OpSlnt)
						fprintf(stdout,"\nERROR:- Pattern Memory Label %s is not defined on line %3.3d in file %s", operand, LineNum, SourceStack[0]);
						ErrCount++;
					return(-1);
				}
				break;
			case PGMSEEK:
				for(i = 0; i < TotalLabels; i++) {
					if(strcmp(operand, Label[i].String) == 0) {
						value = (unsigned)Label[i].Address;
						break;
					}
				}
				if(i >= TotalLabels) {
					if(!OpSlnt)
						fprintf(stdout,"\nERROR:- Jump Address Label %s is not defined on line %3.3d in file %s", operand, LineNum, SourceStack[0]);
					ErrCount++;
					return(-1);
				}
				break;
			case DEFSEEK:
				for(i = 0; i < DEFDATASIZE; i++) {
					if(strcmp(operand, defData[i].String) == 0) {
						value = defData[i].Value;
						break;
					}
				}
				if(i >= DEFDATASIZE) {
					if(!OpSlnt)
						fprintf(stdout,"\nERROR:- Definition for %s is not defined on line %3.3d in file %s", operand, LineNum, SourceStack[0]);
					ErrCount++;
					return(-1);
				}
				break;
		}
	}
	else {
		value = parseValue(operand);
		i = 0;							// To signal a good return value.
	}

// We should mask the data down to the correct precision here
	m = (unsigned long) pow(2, (4 * OpSize));
	if(m <= value) {
		l = value & (m - 1);
		if(!OpSlnt)
			fprintf(stdout,"\nWARNING:-  Value (0x%x) has overflowed precision on line %3.3d in file %s\n           Value has been truncated to 0x%x", value, LineNum, SourceStack[0], l);
		WarnCount++;
		value = l;
	}

// Now convert the value into nibbles and stuff them into the program memory area.

	for (j = OpSize ; j > 1 ;j--) {
		m = (unsigned long)pow(2,4 * (j - 1));
		l = value / m;
		value = value - (l * m);
		Mem[pgmAddr + j] = (unsigned char)l;
	}
	Mem[pgmAddr + j] = (unsigned char)value;

	return((int)i);
}

//
// Extract an unsigned integer from whatever format the input string happens to be in
//
unsigned long parseValue(char *ptr)
{
	int len, i;
	unsigned long value = 0;

	while(*ptr == '0')				// Skip leading zeros.
		ptr++;

	switch(*ptr)
	{
		case 'B':
			ptr++;
			len = strlen(ptr);
			for (i = 0; i < len; i++) {
				if(ptr[i] == '1') 
					value = value + (unsigned)pow(2,len - i - 1);
			}
			break;
		case 'X':
		case 'H':
			sscanf(++ptr,"%X",&value);
			break;
		case 'D':
			sscanf(++ptr,"%d",&value);
			break;
		default:
			if (strlen(ptr) != 0) {
				if (isdigit((int)*ptr))
					sscanf(ptr ,"%d",&value);
				else {
					if(!OpSlnt)
						fprintf(stdout,"\nWARNING:- Definition for %s is possibly missing '#' on line %3.3d in file %s", ptr, LineNum, SourceStack[0]);
					WarnCount++;
				}
			}
			break;
	}
	return(value);
}

//
// Identify the correct opcode from the source and return the equivelent bin representation
//
unsigned char findOpcode(char* Opcode)
{
	if(strcmp(Opcode,"LPP")== 0) 
		return(LPP);
	else if(strcmp(Opcode,"IPP") == 0) 
		return(IPP);
	else if(strcmp(Opcode,"DPP") == 0) 
		return(DPP);
	else if(strcmp(Opcode,"LDA") == 0) 
		return(LDA);
	else if(strcmp(Opcode,"LMR") == 0) 
		return(LMR);
	else if(strcmp(Opcode,"LSR") == 0) 
		return(LSR);
	else if(strcmp(Opcode,"CAL") == 0) 
		return(CAL);
	else if(strcmp(Opcode,"RET") == 0) 
		return(RET);
	else if(strcmp(Opcode,"JCB") == 0) 
		return(JCB);
	else if(strcmp(Opcode,"LRB") == 0) 
		return(LRB);
	else if(strcmp(Opcode,"LPB") == 0) 
		return(LPB);
	else if(strcmp(Opcode,"LPE") == 0) 
		return(LPE);
	else if(strcmp(Opcode,"DMS") == 0) 
		return(DMS);
	else if(strcmp(Opcode,"DUS") == 0) 
		return(DUS);
	else if(strcmp(Opcode,"DSC") == 0) 
		return(DSC);
	else if(strcmp(Opcode,"NOP") == 0) 
		return(NOP);
	else if(strcmp(Opcode,"DEF") == 0) 
		return(DEF);
	else if(strcmp(Opcode,"PAT") == 0) 
		return(PAT);
	else if(strcmp(Opcode,"ORG") == 0) 
		return(ORG);
	else if(strcmp(Opcode,"JMP") == 0) 
		return(JMP);
	else if(strcmp(Opcode,"NOT") == 0) 
		return(NOT);
	else if(strcmp(Opcode,"AND") == 0) 
		return(AND);
	else if(strcmp(Opcode,"OR") == 0) 
		return(OR);
	else if(strcmp(Opcode,"XOR") == 0) 
		return(XOR);	
	else if(strcmp(Opcode,"INC") == 0) 
		return(INC);	
	else
		return(BAD);
}

//
// Output to file in dheDbgFormat (BrdAddress, Address, Value)
// 
void printDheDbg(FILE *fpt, char *src)
{
	unsigned i, j, lowWord, highWord;
	time_t ttime;
	struct tm *stime;
	char ctime[27];

	ttime = time(NULL);			// Fetch the current time (very laborious !!)
	stime = localtime(&ttime);
	strftime(ctime,26,"%d-%b-%Y %H:%M:%S",stime);

	if(MaxPatAddr != 0) {
		if(OpDmp)
			fprintf(stdout,"\n\n# COMMENT - PATTERN SEGMENT IMAGE");	
		fprintf(fpt,"#\n# COMMENT - PATTERN SEGMENT IMAGE\n# This file generated from %s on %s\n#", src, ctime);	
		fprintf(fpt," via %s Version %s\n#", PRODUCTNAME, PRODUCTVERS);	
		for ( i = 0 ;i < MaxPatAddr; i++) {
			highWord = (unsigned)(patData[i].Value / pow(2, 16));
			lowWord = (unsigned)(patData[i].Value - (highWord * pow(2, 16)));
			if(OpDmp)							// Print out the lower 16 bit value first.
				fprintf(stdout,"\n0x0001  0x%4.4x  0x%4.4x",(i * 2) + DHEPATSEGADDR, lowWord);
			fprintf(fpt,"\n0x0001  0x%4.4x  0x%4.4x",(i * 2) + DHEPATSEGADDR, lowWord);
			if(OpDmp)							// Print out the lower 16 bit value first.
				fprintf(stdout,"\n0x0001  0x%4.4x  0x%4.4x",(i * 2) + DHEPATSEGADDR + 1, highWord);
			fprintf(fpt,"\n0x0001  0x%4.4x  0x%4.4x",(i * 2) + DHEPATSEGADDR + 1, highWord);
		} 
		fprintf(fpt,"\n");
	}
			
	if(MaxPgmAddr != 0) {
		if(OpDmp)
			fprintf(stdout,"\n\n# COMMENT - PROGRAM SEGMENT IMAGE");	
		fprintf(fpt,"#\n# COMMENT - PROGRAM SEGMENT IMAGE\n# This file generated from %s on %s\n#",src, ctime);	
		fprintf(fpt," via %s Version %s\n#", PRODUCTNAME, PRODUCTVERS);	
		for ( i = 0 ;i < (unsigned)ceil(pgmAddr/4) + 1; i++)
		{
			if(OpDmp)
				fprintf(stdout,"\n0x0001  0x%4.4x  0x",i + DHEPGMSEGADDR);
			fprintf(fpt,"\n0x0001  0x%4.4x  0x",i + DHEPGMSEGADDR);
			for (j = 4;j > 0;j--) {
				if(OpDmp)
					fprintf(stdout,"%1x",Mem[4 * i + (j - 1)]);
				fprintf(fpt,"%1x",Mem[4 * i + (j - 1)]);
			}
		}
		fprintf(fpt,"\n# END OF PROGRAM SEGMENT\n");
	}
	return;
}

//
// Output to file in ipx configure format (Value Value .... Value, etc.)
// 
void printPpx(FILE *fpt, char *src)
{
	FILE *aps = NULL;
	unsigned i, j, k, highWord, lowWord;
	time_t ttime;
	struct tm *stime;
	char ctime[27];

	ttime = time(NULL);			// Fetch the current time (very laborious !!)
	stime = localtime(&ttime);
	strftime(ctime,26,"%d-%b-%Y %H:%M:%S",stime);

	if(patAddr != 0) {
		if(OpDmp)
			fprintf(stdout,"\n Pattern Memory Segment Data\n");
		fprintf(fpt,"##");
		fprintf(fpt,"\n## SEQUENCER PATTERN MEMORY SEGMENT");
		fprintf(fpt,"\n## (CHECK THE ADDRESS REFERENCE !!)");
		fprintf(fpt,"\n## This file generated from source file %s on %s\n##",src, ctime);
		fprintf(fpt," via %s Version %s\n##", PRODUCTNAME, PRODUCTVERS);	
		fprintf(fpt,"\naddr=mcbSeqPatMem        # pattern memory address\n");
		i = 0;
		while(i < MaxPatAddr) {
			for(k = 0; k < 4 ; k++) {
				highWord = (unsigned)(patData[i + k].Value / pow(2, 16));
				lowWord = (unsigned)(patData[i + k].Value - (highWord * pow(2, 16)));

				if(OpDmp)
					fprintf(stdout,"0x%4.4x 0x%4.4x ",lowWord, highWord);
				fprintf(fpt,"0x%4.4x 0x%4.4x ",lowWord, highWord);
			}
			if(OpDmp)
				fprintf(stdout,"\n");
			fprintf(fpt,"\n");
			i += 4;
		}
		if(OpDmp)
			fprintf(stdout,"\n");
	}
	if(OpDmp)
		fprintf(stdout,"\n Program Memory Segment Data\n");
	fprintf(fpt,"##");
	fprintf(fpt,"\n## SEQUENCER PROGRAM MEMORY SEGMENT");
	fprintf(fpt,"\n## (CHECK THE ADDRESS REFERENCE !!)");
	fprintf(fpt,"\n## This file generated from source file %s on %s\n##",src, ctime);
	fprintf(fpt," via %s Version %s\n##", PRODUCTNAME, PRODUCTVERS);	
	fprintf(fpt,"\naddr=mcbSeqPgmMem        # program memory address\n");
	i = 0;
	while(i < (unsigned)ceil(pgmAddr/4) + 1)
	{
		for(k = 0; k < 8; k++) {
			if(OpDmp)
				fprintf(stdout,"0x");
			fprintf(fpt,"0x");
			for (j = 4;j > 0;j--) {
				if(OpDmp)
					fprintf(stdout,"%x",Mem[4 * i + (j - 1)]);
				fprintf(fpt,"%x",Mem[4 * i + (j - 1)]);
			}
			i++;
			if(OpDmp)
				fprintf(stdout," ");
			fprintf(fpt," ");
		}
		if(OpDmp)
			fprintf(stdout,"\n");
		fprintf(fpt,"\n");
	} 
	
	return;
}


int PushSource(char *fname)
{
	int q;

	if (StackCntr < (MAX_INCLUDES - 1)) {
		for(q = MAX_INCLUDES-2;q >= 0; q--) {
			strcpy(SourceStack[q+1],SourceStack[q]);
			LineNumStack[q+1] = LineNumStack[q];
		}
		strcpy(SourceStack[0],fname);
		LineNumStack[0] = LineNum;
		LineNum = 0;
		StackCntr++;
		return(StackCntr);
	}
	else {
		return(-1);
	}
}

int PopSource(void)
{
	int q;

	if (StackCntr > 0) {
		LineNum = LineNumStack[0];
		for(q = 1;q < MAX_INCLUDES; q++) {
			strcpy(SourceStack[q-1],SourceStack[q]);
			LineNumStack[q-1] = LineNumStack[q];
		}
		StackCntr--;
		return(StackCntr);
	}
	else {
		return(-1);
	}
}

//
// End of source asm5.cpp
//
